from typing import Any, List, Optional
from datetime import datetime
import logging

from pytweetql._logging import logging_config
from pytweetql._typing import ResponseKey

logger = logging.getLogger(__name__)
logging_config(logger=logger)

def verify_boolean(boolean: Any) -> Optional[bool]:
    """
    Verify expected boolean value type. If not a boolean will return None.

    Args:
        boolean (Any): Expected boolean value.

    Returns:
        bool, optional: The boolean value.
    """
    if isinstance(boolean, bool):
        return boolean
    elif isinstance(boolean, str):
        if boolean.lower() == 'true':
            return True
        elif boolean.lower() == 'false':
            return False


def verify_integer(integer: Any) -> Optional[int]:
    """
    Verify expected integer value type. If not an integer will return None.

    Args:
        integer (Any): Expected integer value.

    Returns:
        int, optional: The integer value.
    """
    if integer is None:
        return
    elif isinstance(integer, int):
        return integer
    elif isinstance(integer, float):
        number_int = int(integer)
        if number_int == integer:
            return number_int
    elif isinstance(integer, str):
        try:
            return int(integer)
        except ValueError:
            return


def verify_datetime(created: Any) -> Any:
    """
    Verfiy format of created date provided in response.

    Args:
        created (Any): The string representation of the created date field.

    Returns:
        Any: The converted UTC string datetime into an ISO format datetime.
    """
    try:
        return datetime.strptime(created, "%a %b %d %H:%M:%S %z %Y").isoformat()
    except ValueError:
        logger.warning(
            f'Incorrect date format specified: {created}, date formatting being skipped'
        )
        return created


def extract_dicts_from_list(source: list) -> List[dict]:
    """
    Recursive function to extract the top-level dictionaries into one list
    from nested lists.

    Args:
        source (list): A list of dictionaries.

    Returns:
        List[dict]: A list with the found dictionary or an empty list.
    """
    def helper(source: list, target: list) -> list:
        if isinstance(source, dict):
            target.append(source)
            
        if isinstance(source, list):
            for item in source:
                target.extend(helper(item, []))
        return target
    
    return helper(source=source, target=[])


def empty_dictionary(source: ResponseKey) -> bool:
    """
    A recursive function to determine if dictionary is empty.

    Args:
        source (ResponseKey): A list or dictionary.

    Returns:
        bool: Whether the object is empty.
    """
    if isinstance(source, dict):
        return all(empty_dictionary(value) for _, value in source.items())
    elif isinstance(source, list):
        return all(empty_dictionary(element) for element in source)
    else:
        return not source


from dataclasses import dataclass
from typing import List

@dataclass
class Error:
    """
    A parsing error.

    Attributes:
        code (str): Code associated with the error.
        message (str): Message associated with the error.
        api_code (int, optional): An API code, if the error was generated from the API.
    """
    code: str
    message: str
    api_code: int = None


def is_api_error_specific(code: int, errors: List[Error]) -> bool:
    """Detect if a specific error has occurred, using the API code."""
    return any(error.api_code == code for error in errors)


def is_api_error(error: Error) -> bool:
    """Detect if error is an API error."""
    return error.api_code is not None


def detect_api_errors(errors: List[Error]) -> List[Error]:
    """Get all API errors associated with response."""
    return [error for error in errors if is_api_error(error=error)]


def generate_api_error(code: int, message: str) -> Error:
    """Generate dynamic API error from message."""
    return Error(code='ERR_007', message=message, api_code=code)


ERROR_NONE = Error(code='ERR_001', message='Response is null')
ERROR_FORMAT = Error(code='ERR_002', message='Unknown format')
ERROR_JSON = Error(code='ERR_003', message='Invalid JSON')
ERROR_PARSER = Error(code='ERR_004', message='Invalid parser')
ERROR_EMPTY = Error(code='ERR_005', message='Response is empty')
ERROR_API_UNKNOWN = Error(code='ERR_006', message='Unknown API error')


from typing import List

from pytweetql.validation.validation import DirectPathValidation
from pytweetql._utils._data_structures import APIErrorInfo
from pytweetql.validation._base_validation import error_check_output
from pytweetql._typing import (
    APIResponse,
    Schema
)

class APIError:
    """
    Parsing for an individual API error.

    Args:
        code (int): An integer code associated with the error.
        message (str): The message which describes the error.
    """
    def __init__(self, code: int, message: str):
        self._code = code
        self._message = message

        self._api_error = self._parse_error()

    def _parse_error(self) -> APIErrorInfo:
        """
        Parse API error into structured format.
    
        Returns:
            APIErrorInfo: The dataclass which holds all relevant error detail.
        """
        return APIErrorInfo(code=self._code, message=self._message)

    @property
    def api_error(self) -> APIErrorInfo:
        """The entire APIErrorInfo dataclass."""
        return self._api_error

    @property
    def message(self) -> str:
        """The message which describes the error."""
        return self._message

    @property
    def code(self) -> int:
        """An integer code associated with the error."""
        return self._code

class APIErrors(DirectPathValidation):
    """
    Parsing for an error API response.

    Args:
        response (APIResponse): The response from a Twitter API.
        schema (Schema): The schema used to validate the API response.
    """
    def __init__(
        self,
        response: APIResponse,
        schema: Schema
    ):
        super().__init__(response=response, do_errors=True)
        self._schema = schema

        self._api_errors = self._parse_api_errors()

    @property
    def api_errors(self) -> List[APIError]:
        """Returns all the parsed API errors."""
        return self._api_errors

    @property
    def num_api_errors(self) -> int:
        """The number of API errors parsed in response."""
        return len(self._api_errors)
    
    @error_check_output
    def _parse_api_errors(self) -> List[APIError]:
        """
        Parse each individual error detail from response.

        Returns:
            List[APIError]: A list of Error classes, one for each API error detected.
        """
        return [
            APIError(**entry) for entry in self.extract_objects(
                schema=self._schema
            )
        ]


from pytweetql._utils._utils import (
    verify_boolean,
    verify_datetime,
    verify_integer
)

class BaseTweet:
    """
    Raw parsing for an individual tweet.
    
    Args:
        user (dict): The raw user section in each tweet response.
        user_info (dict): The raw user info section in each tweet response.
        tweet (dict): The raw tweet section in each tweet response.
        source (str): The raw source section in each tweet response.
    """
    def __init__(self, user: dict, user_info: dict, tweet: dict, source: str):
        self._user = user
        self._user_info = user_info
        self._tweet = tweet
        self._source = source

    @property
    def _user_id(self) -> str:
        """The user ID of the Twitter account that posted the tweet."""
        return self._user.get('rest_id')

    @property
    def _user_name(self) -> str:
        """The user name of the Twitter account that posted the tweet."""
        return self._user_info.get('name')
        
    @property
    def _user_screen_name(self) -> str:
        """The user screen name of the Twitter account that posted the tweet."""
        return self._user_info.get('screen_name')

    @property
    def _tweet_id(self) -> str:
        """The tweet ID."""
        return self._tweet.get('id_str')

    @property
    def _created_date(self) -> str:
        """The UTC date the tweet was created."""
        return verify_datetime(created=self._tweet.get('created_at'))
        
    @property
    def _content(self) -> str:
        """The text content of the tweet."""
        return self._tweet.get('full_text')

    @property
    def _language(self) -> str:
        """The language of the text content."""
        return self._tweet.get('lang')
    
    @property
    def _tweet_url(self) -> str:
        """The URL of the tweet."""
        return f'https://twitter.com/{self._user_name}/status/{self._tweet_id}'
    
    @property
    def _quote_count(self) -> int:
        """The number of times the tweet has been quoted."""
        return verify_integer(integer=self._tweet.get('quote_count'))
        
    @property
    def _reply_count(self) -> int:
        """The number of replies on the tweet."""
        return verify_integer(integer=self._tweet.get('reply_count'))
        
    @property
    def _retweet_count(self) -> int:
        """The number of times the tweet has been retweeted."""
        return verify_integer(integer=self._tweet.get('retweet_count'))
        
    @property
    def _is_quote(self) -> bool:
        """Boolean indicating whether it is a quoted tweet."""
        return verify_boolean(boolean=self._tweet.get('is_quote_status'))
        
    @property
    def _is_retweet(self) -> bool:
        """Boolean indicating whether it is a retweet."""
        return verify_boolean(boolean=self._tweet.get('retweeted'))


class BaseUser:
    """
    Raw parsing for an individual user.

    Args:
        user (str): The raw user ID in each user response.
        user_info (dict): The raw user info section in each user response.
    """
    def __init__(self, user: str, user_info: dict):
        self._user = user
        self._user_info = user_info

    @property
    def _user_id(self) -> str:
        """The user ID of the Twitter account."""
        return self._user
    
    @property
    def _user_name(self) -> str:
        """The user name of the Twitter account."""
        return self._user_info.get('name')
        
    @property
    def _user_screen_name(self) -> str:
        """The user screen name of the Twitter account."""
        return self._user_info.get('screen_name')
    
    @property
    def _profile_description(self) -> str:
        """The user profile description."""
        return self._user_info.get('description')

    @property
    def _created_date(self) -> str:
        """The UTC date the user profile was created."""
        return verify_datetime(created=self._user_info.get('created_at'))

    @property
    def _location(self) -> str:
        """The Location of user."""
        return self._user_info.get('location')

    @property
    def _favourites_count(self) -> int:
        """The number of favorites."""
        return verify_integer(integer=self._user_info.get('favourites_count'))

    @property
    def _followers_count(self) -> int:
        """The number of followers."""
        return verify_integer(integer=self._user_info.get('followers_count'))

    @property
    def _statuses_count(self) -> int:
        """The number of statuses."""
        return verify_integer(integer=self._user_info.get('statuses_count'))

    @property
    def _is_verified(self) -> bool:
        """Boolean indicating whether the user is verified."""
        return verify_boolean(boolean=self._user_info.get('verified'))


from typing import Any, Generator, List, Tuple
import copy

from pytweetql.errors import *
from pytweetql.validation._base_validation import BaseValidation
from pytweetql.validation._nodes import nodes_error_api
from pytweetql._typing import (
    APIResponse, 
    Schema
)
from pytweetql._utils._utils import (
    empty_dictionary,
    extract_dicts_from_list
)
from pytweetql.validation._node_path import (
    NodePath,
    PathNode
)

def _derive_node_path(schema: dict) -> NodePath:
    """
    Create the node path from a sceham dictionary.

    Args:
        schema (dict): The dictionary schema used to validate the API response.

    Returns:
        NodePath: The node path as a linked list.
    """
    return NodePath(schema=schema)


def _load_schema(schema: Schema) -> Tuple[dict, dict]:
    """
    Validate and load schema specification.
    
    Args:
        schema (Schema): The schema used to validate the API response.

    Returns:
        Tuple[dict, dict]: A tuple of the two Schema components.
    """
    if 'entry' not in schema or 'objects' not in schema:
        raise TypeError("Schema must only have 'entry' and 'objects' keys")
    return schema['entry'], schema['objects']


class DirectPathValidation(BaseValidation):
    """
    Functionality to parse and validate a direct path to dictionaries in a JSON.

    Args:
        response (APIResponse): The response from a Twitter API.
        do_errors (bool): Whether to validate the error format (or main format).
    """
    def __init__(self, response: APIResponse, do_errors: bool = False):
        super().__init__(response=response)

        if not do_errors:
            # Detect and extract any API errors in response
            self._extract_api_errors()

            # Validate response and check if result is empty
            self._validate_response()
        if empty_dictionary(source=self.response):
            self._error(error=ERROR_EMPTY)
        
    def _extract_api_errors(self) -> None:
        """Extract all API errors from response."""
        errors = self.extract_objects(schema=nodes_error_api)
        for error in errors:
            self._error(
                error=generate_api_error(**error)
            )

    def _if_been_list(
        self, 
        response: List[dict], 
        current_node: PathNode
    ) -> list:
        """
        Iterate through a list object to find the next key in the path.

        Args:
            responses (List[dict]): A list of response dictionaries.
            current_node (PathNode): The active node in linked list.

        Returns:
            list: A list of objects found.
        """
        key_search = []
        for item in response:
            if isinstance(item, dict):
                list_value = item.get(current_node.key)
                if list_value is not None:
                    key_search.append(list_value)
        return key_search

    def extract_objects(self, schema: Schema) -> Generator[Any, None, None]:
        """
        Extract all relevant objects from the response dictionary.

        Args:
            schema (Schema): The schema used to validate the API response.

        Returns:
            Generator[Any, None, None]: Generator that yields any data type.
        """
        # Load in schema
        schema_entry, schema_objects = _load_schema(schema=schema)

        # Derive entry node path
        node_path_entry = _derive_node_path(schema=schema_entry)

        # Validate schema
        entries = self._validate_schema(
            node_path=node_path_entry,
            responses=self.response
        )
        for entry in entries:
            entry_results = {}
            for arg, schema in schema_objects.items():
                results = self._validate_schema(
                    node_path=_derive_node_path(schema=schema),
                    responses=[entry]
                )
                for result in results:
                    entry_results.update({arg: result})
            yield entry_results

    def _validate_schema(
        self,
        node_path: NodePath,
        responses: List[dict]
    ) -> Generator[Any, None, None]:
        """
        Validate schema of response by navigating through linked list key path.
        
        Args:
            responses (List[dict]): A list of response dictionaries.
            node_path (NodePath): A linked list of PathNodes from schema.

        Returns:
            Generator[Any, None, None]: Generator that yields any data type.
        """
        for resp in responses:
            current_node = node_path.head
            while current_node is not None:

                # Find the current node key in response
                obj = None
                if node_path.been_list:
                    obj = self._if_been_list(
                        response=resp, 
                        current_node=current_node
                    )
                elif isinstance(resp, dict):
                    obj = resp.get(current_node.key)

                # Match type of found value to expected type
                if node_path.isinstance_of_type(
                    obj=obj, 
                    node=current_node
                ):
                    resp = copy.copy(obj)
                    previouse_node = current_node.prev

                    # If value of previous node was a list,
                    # then extract all objects from the result
                    if (
                        previouse_node is not None and 
                        previouse_node.value_type == 'list'
                    ):
                        resp = extract_dicts_from_list(source=resp) 

                    # If current node is a result, then add to entries and break
                    if not current_node.next:
                        if isinstance(resp, list):
                            for record in resp:
                                yield record
                        else:
                            yield resp
                        break
                    else:
                        current_node = current_node.next
                else:
                    break


from dataclasses import dataclass

@dataclass(frozen=True)
class APIErrorInfo:
    """
    A request error generated by the API.

    Attributes:
        code (int): An integer code associated with the error.
        message (str): The message which describes the error.
    """
    code: int
    message: str


@dataclass(frozen=True)
class ListInfo:
    """
    Info pulled from the Twitter list profile.

    Attributes:
        name (str): The list name.
        description (str): The list description.
        list_id (str): The list ID.
        mode (str): The mode of the list (public or private).
        member_count (int): The number of members in the list.
        is_following (bool): Boolean indicating whether the user is following the list.
    """
    name: str
    description: str
    list_id: str
    mode: str
    member_count: int
    is_following: bool


@dataclass(frozen=True)
class UserInfo:
    """
    Info pulled from a user profile.

    Attributes:
        user_id (str): The user ID of the Twitter account that posted the tweet.
        user_name (str): The user name of the Twitter account that posted the tweet.
        user_screen_name (str): The user screen name of the Twitter account that posted the tweet.
        profile_description (str): The user profile description.
        created (str): The UTC date the user profile was created.
        location (str): The Location of user.
        favourites_count (int): The number of favorites.
        followers_count (int): The number of followers.
        statuses_count (int): The number of statuses.
        is_verified (bool): Boolean indicating whether the user is verified.
    """
    user_id: str
    user_name: str
    user_screen_name: str
    profile_description: str
    created: str
    location: str
    favourites_count: int
    followers_count: int
    statuses_count: int
    is_verified: bool


@dataclass(frozen=True)
class TweetInfo:
    """
    Info pulled from a tweet.

    Attributes:
        user_id (str): The user ID of the Twitter account that posted the tweet.
        user_name (str): The user name of the Twitter account that posted the tweet.
        user_screen_name (str): The user screen name of the Twitter account that posted the tweet.
        tweet_id (str): The tweet ID.
        created (str): The UTC date the tweet was created.
        content (str): The text content of the tweet.
        language (str): The language of the text content.
        tweet_url (str): The URL of the tweet.
        quote_count (int): The number of times the tweet has been quoted.
        reply_count (int): The number of replies on the tweet.
        retweet_count (int): The number of times the tweet has been retweeted.
        is_quote (bool): Boolean indicating whether it is a quoted tweet.
        is_retweet (bool): Boolean indicating whether it is a retweet.
    """
    user_id: str
    user_name: str
    user_screen_name: str
    tweet_id: str
    created: str
    content: str
    language: str
    tweet_url: str
    quote_count: int
    reply_count: int
    retweet_count: int
    is_quote: bool
    is_retweet: bool


from pytweetql.response.twitterlist import TwitterLists
from pytweetql.validation._nodes import *
from pytweetql._typing import APIResponse
from pytweetql.response.user import Users
from pytweetql.response.tweet import Tweets
from pytweetql.response.api_error import APIErrors

def parse_api_errors(response: APIResponse) -> APIErrors:
    """
    Parse any API errors found in response.
    """
    return APIErrors(
        response=response,
        schema=nodes_error_api
    )


def parse_create_list(response: APIResponse) -> TwitterLists:
    """
    Parse Twitter list data from the CreateList endpoint.
    """
    return TwitterLists(
        response=response,
        schema=nodes_list_create,
        endpoint='CreateList'
    )


def parse_following(response: APIResponse) -> Users:
    """
    Parse user data from the Following endpoint.
    """
    return Users(
        response=response,
        schema=nodes_following,
        endpoint='Following'
    )


def parse_list_remove_member(response: APIResponse) -> Users:
    """
    Parse user data from the ListRemoveMember endpoint.
    """
    return Users(
        response=response,
        schema=nodes_list_remove_member,
        endpoint='ListRemoveMember'
    )


def parse_list_add_member(response: APIResponse) -> Users:
    """
    Parse user data from the ListAddMember endpoint.
    """
    return Users(
        response=response,
        schema=nodes_list_add_member,
        endpoint='ListAddMember'
    )


def parse_user_by_id(response: APIResponse) -> Users:
    """
    Parse user data from the UserByRestId endpoint.
    """
    return Users(
        response=response,
        schema=nodes_user_by_rest_id,
        endpoint='UserByRestId'
    )


def parse_users_by_ids(response: APIResponse) -> Users:
    """
    Parse user data from the UsersByRestIds endpoint.
    """
    return Users(
        response=response,
        schema=nodes_users_by_rest_ids,
        endpoint='UsersByRestIds'
    )


def parse_list_members(response: APIResponse) -> Users:
    """
    Parse user data from the ListMembers endpoint.
    """
    return Users(
        response=response,
        schema=nodes_list_members,
        endpoint='ListMembers'
    )


def parse_users_by_screen_name(response: APIResponse) -> Users:
    """
    Parse user data from the UserByScreenName endpoint.
    """
    return Users(
        response=response,
        schema=nodes_user_by_screen_name,
        endpoint='UserByScreenName'
    )


def parse_tweet_result_by_id(
    response: APIResponse, 
    remove_promotions: bool = True
) -> Tweets:
    """
    Parse tweet data from the TweetResultByRestId endpoint.
    """
    return Tweets(
        response=response,
        schema=nodes_tweet_result_by_id,
        remove_promotions=remove_promotions,
        endpoint='TweetResultByRestId'
    )


def parse_user_tweets(
    response: APIResponse, 
    remove_promotions: bool = True
) -> Tweets:
    """
    Parse tweet data from the UserTweets endpoint.
    """
    return Tweets(
        response=response,
        schema=nodes_user_tweets,
        remove_promotions=remove_promotions,
        endpoint='UserTweets'
    )


def parse_create_tweet(
    response: APIResponse, 
    remove_promotions: bool = True
) -> Tweets:
    """
    Parse tweet data from the CreateTweet endpoint.
    """
    return Tweets(
        response=response,
        schema=nodes_create_tweet,
        remove_promotions=remove_promotions,
        endpoint='CreateTweet'
    )


def parse_list_latest_tweets(
    response: APIResponse, 
    remove_promotions: bool = True
) -> Tweets:
    """
    Parse tweet data from the ListLatestTweetsTimeline endpoint.
    """
    return Tweets(
        response=response,
        schema=nodes_list_latest_tweets,
        remove_promotions=remove_promotions,
        endpoint='ListLatestTweetsTimeline'
    )


from typing import Any, Literal, Optional

_TYPE_MAPPING = {
    'list': list,
    'dict': dict,
    'int': int,
    'str': str,
    'bool': bool
}

class PathNode:
    """
    An individual node in the path.
    """
    def __init__(
        self, 
        key: Any,
        value_type: Literal['str', 'bool', 'list', 'int', 'dict']
    ):
        self.key = key
        self.value_type = value_type
        self.prev: Optional['PathNode'] = None
        self.next: Optional['PathNode'] = None


class NodePath:
    """
    A linked list of PathNodes from schema.

    Args:
        schema (dict): The dictionary schema used to validate the API response.
    """
    def __init__(self, schema: dict):
        self.head = None
        self.been_list = False
        
        self._construct_linked_list(schema=schema)

    def _append_node_to_bottom(self, node: PathNode) -> None:
        """
        Add a node to bottom of the linked list.

        Args:
            node (PathNode): The node to add to list.
        """
        if self.head is None:
            self.head = node
            return

        node_last = self.head
        while node_last.next:
            node_last = node_last.next

        node.prev = node_last
        node_last.next = node

    def _construct_linked_list(self, schema: dict) -> None:
        """
        Construct the doubly linked list.

        Args:
            schema (dict): The dictionary schema used to validate the API response.
        """
        for key, value in schema.items():
            if 'type' not in value:
                raise ValueError('key type needs to be in schema')

            node = PathNode(key=key, value_type=value['type'])
            self._append_node_to_bottom(node=node)
            schema = schema[key]
            
        if 'children' in schema:
            self._construct_linked_list(schema['children'])
    
    def isinstance_of_type(self, obj: Any, node: PathNode) -> bool:
        """
        Validate type of found object in response.

        Args:
            obj (Any): The object found in the step search.
            node (PathNode): A node in the validation schema.

        Returns:
            bool: Whether the object matches the expected type.
        """
        expected_type = _TYPE_MAPPING.get(node.value_type)
        if expected_type is None:
            raise ValueError(
                'Argument type_value specified for node is invalid'
            )

        if self.been_list:
            return all(isinstance(var, expected_type) for var in obj)
        else:
            if expected_type == list:
                self.been_list = True
            return isinstance(obj, expected_type)


from pytweetql.constants import GraphQLCodes
from pytweetql import parsing
from pytweetql.response.api_error import APIError
from pytweetql.errors import (
    detect_api_errors,
    Error,
    is_api_error_specific,
    is_api_error
)
from pytweetql.response.twitterlist import (
    TwitterList, 
    TwitterLists
)
from pytweetql.response.tweet import (
    Tweet, 
    Tweets
)
from pytweetql.response.user import (
    User, 
    Users
)

__version__ = '0.5.0'
__all__ = [
    'APIError',
    'detect_api_errors',
    'Error',
    'GraphQLCodes',
    'is_api_error_specific',
    'is_api_error',
    'parsing',
    'Tweet', 
    'Tweets',
    'TwitterList', 
    'TwitterLists',
    'User', 
    'Users'
]


import json
from typing import List

from pytweetql.errors import *
from pytweetql._utils._utils import extract_dicts_from_list
from pytweetql._typing import APIResponse

def error_check_output(func) -> None:
    """
    Decorator to add invalid parser error.
    """
    def wrapper(self, *args, **kwargs):
        data = func(self, *args, **kwargs)
        if not data and not self.errors:
            self._error(error=ERROR_PARSER)
        return data
    return wrapper


class BaseStatus:
    """
    Base methods and functionality for managing response status.
    """
    def __init__(self):
        self._status_code = 200
        self._errors: List[Error] = []

    def _error(self, error: Error) -> None:
        """Append error to list of errors."""
        self._errors.append(error)
        if self._status_code == 200:
            self._status_code = 400

    @property
    def errors(self) -> List[Error]:
        """Returns all errors."""
        return self._errors
    
    @property
    def status_code(self) -> str:
        """Return the current status code."""
        return self._status_code


class BaseValidation(BaseStatus):
    """
    Functionality to run validation on response.

    Args:
        response (APIResponse): The response from a Twitter API.
    """
    def __init__(self, response: APIResponse):
        super().__init__()
        self._validate_input_type(response=response)

    @property
    def response(self) -> dict:
        """Return the parsed response."""
        return self._response
    
    @response.setter
    def response(self, response: APIResponse) -> None:
        """Set a new parsed response."""
        self._response = response

    def _validate_input_type(self, response: APIResponse) -> None:
        """
        Validate the response input. Ensure that input is a list or dict. If JSON,
        load and convert.

        Args:
            response (APIResponse): The response from a Twitter API.
        """
        if response is None:
            self._error(error=ERROR_NONE)

        if isinstance(response, str):
            try:
                response = json.loads(response)
            except json.JSONDecodeError:
                self._error(error=ERROR_JSON)
        if not isinstance(response, (dict, list)):
            self._error(error=ERROR_FORMAT)
        else:
            self.response = response
    
    def _validate_response(self) -> None:
        """Validate and ensure the response is from GraphQL."""
        response = self.response.copy()
        
        # If response is a dictionary, convert to list for easy manipulation
        if isinstance(response, dict):
            response = [response]

        if isinstance(response, list):
            _response = []
            response_extracted = extract_dicts_from_list(source=response)

            for item in response_extracted:
                data_value = item.get('data')
                if isinstance(data_value, dict):
                    _response.append(data_value)

            if _response:
                self.response = _response
                return
            else:
                self._error(error=ERROR_FORMAT)


from typing import List

from pytweetql.validation.validation import DirectPathValidation
from pytweetql._utils._data_structures import ListInfo
from pytweetql.validation._base_validation import error_check_output
from pytweetql._typing import Schema

class TwitterList:
    """
    Parsing for an individual Twitter list.
    
    Args:
        name (str): The list name.
        description (str): The list description.
        list_id (str): The list ID.
        mode (str): The mode of the list (public or private).
        member_count (int): The number of members in the list.        
        is_following (bool): Boolean indicating whether the user is following the list.
    """
    def __init__(
        self,
        name: str,
        description: str,
        is_following: bool,
        list_id: str,
        member_count: int,
        mode: str
    ):
        self._name = name
        self._description = description
        self._is_following = is_following
        self._list_id = list_id
        self._member_count = member_count
        self._mode = mode

        self._list = self._parse_list()

    def _parse_list(self) -> ListInfo:
        """
        Parse Twitter list info into structured format.

        Returns:
            ListInfo: The dataclass which holds all relevant Twitter list detail.
        """
        return ListInfo(
            name=self._name,
            description=self._description,
            list_id=self._list_id,
            mode=self._mode,
            member_count=self._member_count,
            is_following=self._is_following
        )
    
    @property
    def twitter_list(self) -> ListInfo:
        """The entire ListInfo dataclass."""
        return self._list
    
    @property
    def name(self) -> str:
        """The list name."""
        return self._list.name
    
    @property
    def description(self) -> str:
        """The list description."""
        return self._list.description
    
    @property
    def list_id(self) -> str:
        """The list ID."""
        return self._list.list_id
    
    @property
    def mode(self) -> str:
        """Whether the list is private or public."""
        return self._list.mode

    @property
    def member_count(self) -> int:
        """The number of members in the list."""
        return self._list.member_count
    
    @property
    def is_following(self) -> bool:
        """Boolean indicating whether the user is following the list."""
        return self._list.is_following


class TwitterLists(DirectPathValidation):
    """
    Parsing for a Twitter list API response.

    Args:
        response (APIResponse): The response from a Twitter API.
        schema (Schema): The schema used to validate the API response.
        endpoint (str): The name the the GraphQL endpoint.
    """
    def __init__(
        self,
        response: List[dict], 
        schema: Schema, 
        endpoint: str
    ):
        super().__init__(response=response)
        self._schema = schema
        self.endpoint = endpoint

        self._lists = self._parse_lists()
    
    @property
    def twitter_lists(self) -> List[TwitterList]:
        """Returns all the parsed Twitter lists."""
        return self._lists

    @property
    def num_lists(self) -> int:
        """The number of Twitter lists parsed in response."""
        return len(self._lists)

    @error_check_output
    def _parse_lists(self) -> List[TwitterList]:
        """
        Parse each individual Twitter list detail from response.

        Returns:
            List[TwitterList]: A list of TwitterList classes, one for each list detected.
        """
        return [
            TwitterList(**entry) for entry in self.extract_objects(
                schema=self._schema
            )
        ]


from typing import List

from pytweetql.response._base_response import BaseTweet
from pytweetql.validation.validation import DirectPathValidation
from pytweetql._utils._data_structures import TweetInfo
from pytweetql.validation._base_validation import error_check_output
from pytweetql._typing import Schema

class Tweet(BaseTweet):
    """
    Parsing for an individual tweet.
    
    Args:
        user (dict): The raw user section in each tweet response.
        user_info (dict): The raw user info section in each tweet response.
        tweet (dict): The raw tweet section in each tweet response.
        source (str): The raw source section in each tweet response.
    """
    def __init__(
        self, 
        user: dict, 
        user_info: dict, 
        tweet: dict, 
        source: str
    ):
        super().__init__(
            user=user,
            user_info=user_info, 
            tweet=tweet, 
            source=source
        )

        self._tweet = self._parse_tweet()

    def _parse_tweet(self) -> TweetInfo:
        """
        Parse tweet info into structured format.

        Returns:
            TweetInfo: The dataclass which holds all relevant tweet detail.
        """
        return TweetInfo(
            user_id=self._user_id,
            user_name=self._user_name,
            user_screen_name=self._user_screen_name,
            tweet_id=self._tweet_id,
            created=self._created_date,
            content=self._content,
            language=self._language,
            tweet_url=self._tweet_url,
            is_quote=self._is_quote,
            is_retweet=self._is_retweet,
            quote_count=self._quote_count,
            reply_count=self._reply_count,
            retweet_count=self._retweet_count
        )

    @property
    def tweet(self) -> TweetInfo:
        """The entire TweetInfo dataclass."""
        return self._tweet

    @property
    def user_id(self) -> str:
        """The user ID of the Twitter account that posted the tweet."""
        return self._tweet.user_id
    
    @property
    def user_name(self) -> str:
        """The user name of the Twitter account that posted the tweet."""
        return self._tweet.user_name
    
    @property
    def user_screen_name(self) -> str:
        """The user screen name of the Twitter account that posted the tweet."""
        return self._tweet.user_screen_name
    
    @property
    def tweet_id(self) -> str:
        """The tweet ID."""
        return self._tweet.tweet_id
    
    @property
    def created_date(self) -> str:
        """The UTC date the tweet was created."""
        return self._tweet.created
    
    @property
    def content(self) -> str:
        """The text content of the tweet."""
        return self._tweet.content
    
    @property
    def language(self) -> str:
        """The language of the text content."""
        return self._tweet.language
    
    @property
    def tweet_url(self) -> str:
        """The URL of the tweet."""
        return self._tweet.tweet_url
    
    @property
    def quote_count(self) -> int:
        """The number of times the tweet has been quoted."""
        return self._tweet.quote_count
    
    @property
    def reply_count(self) -> int:
        """The number of replies on the tweet."""
        return self._tweet.reply_count
    
    @property
    def retweet_count(self) -> int:
        """The number of times the tweet has been retweeted."""
        return self._tweet.retweet_count
    
    @property
    def is_quote(self) -> bool:
        """Boolean indicating whether it is a quoted tweet."""
        return self._tweet.is_quote
    
    @property
    def is_retweet(self) -> bool:
        """Boolean indicating whether it is a retweet."""
        return self._tweet.is_retweet


class Tweets(DirectPathValidation):
    """
    Parsing for a tweet API response.

    Args:
        response (APIResponse): The response from a Twitter API.
        schema (Schema): The schema used to validate the API response.
        remove_promotions (bool): Whether to remove promoted tweets from parsing.
        endpoint (str): The name the the GraphQL endpoint.
    """
    def __init__(
        self,
        response: List[dict],
        schema: Schema,
        remove_promotions: bool,
        endpoint: str
    ):
        super().__init__(response=response)
        self._schema = schema
        self._remove_promotions = remove_promotions
        self.endpoint = endpoint

        self._tweets = self._parse_tweets()

    @property
    def tweets(self) -> List[Tweet]:
        """Returns all the parsed tweets."""
        return self._tweets

    @property
    def num_tweets(self) -> int:
        """The number of tweets parsed in response."""
        return len(self._tweets)
    
    @error_check_output
    def _parse_tweets(self) -> List[Tweet]:
        """
        Parse each individual tweet detail from response.

        Returns:
            List[Tweet]: A list of Tweet classes, one for each tweet detected.
        """
        return [
            Tweet(**entry) for entry in self.extract_objects(
                schema=self._schema
            )
        ]


from typing import List

from pytweetql.response._base_response import BaseUser
from pytweetql.validation.validation import DirectPathValidation
from pytweetql._utils._data_structures import UserInfo
from pytweetql.validation._base_validation import error_check_output
from pytweetql._typing import Schema

class User(BaseUser):
    """
    Parsing for an individual tweet.
    
    Args:
        user (str): The raw user ID in each user response.
        user_info (dict): The raw user info section in each user response.
    """
    def __init__(self, user: dict, user_info: dict):
        super().__init__(user=user, user_info=user_info)

        self._user = self._parse_user()

    def _parse_user(self) -> UserInfo:
        """
        Parse user info into structured format.

        Returns:
            UserInfo: The dataclass which holds all relevant user detail.
        """
        return UserInfo(
            user_id=self._user_id,
            user_name=self._user_name,
            user_screen_name=self._user_screen_name,
            profile_description=self._profile_description,
            created=self._created_date,
            location=self._location,
            favourites_count=self._favourites_count,
            followers_count=self._followers_count,
            statuses_count=self._statuses_count,
            is_verified=self._is_verified
        )
    
    @property
    def user(self) -> UserInfo:
        """The entire UserInfo dataclass."""
        return self._user
    
    @property
    def user_id(self) -> str:
        """The user ID of the Twitter account."""
        return self._user.user_id
    
    @property
    def user_name(self) -> str:
        """The user name of the Twitter account."""
        return self._user.user_name
        
    @property
    def user_screen_name(self) -> str:
        """The user screen name of the Twitter account."""
        return self._user.user_screen_name
    
    @property
    def profile_description(self) -> str:
        """The user profile description."""
        return self._user.profile_description

    @property
    def created_date(self) -> str:
        """The UTC date the user profile was created."""
        return self._user.created

    @property
    def location(self) -> str:
        """The Location of user."""
        return self._user.location

    @property
    def favourites_count(self) -> int:
        """The number of favorites."""
        return self._user.favourites_count

    @property
    def followers_count(self) -> int:
        """The number of followers."""
        return self._user.followers_count

    @property
    def statuses_count(self) -> int:
        """The number of statuses."""
        return self._user.statuses_count

    @property
    def is_verified(self) -> bool:
        """Boolean indicating whether the user is verified."""
        return self._user.is_verified


class Users(DirectPathValidation):
    """
    Parsing for a user API response.

    Args:
        response (APIResponse): The response from a Twitter API.
        schema (Schema): The schema used to validate the API response.
        endpoint (str): The name the the GraphQL endpoint.
    """
    def __init__(
        self,
        response: List[dict], 
        schema: Schema, 
        endpoint: str
    ):
        super().__init__(response=response)
        self._schema = schema
        self.endpoint = endpoint

        self._users = self._parse_users()

    @property
    def users(self) -> List[User]:
        """Returns all the parsed users."""
        return self._users

    @property
    def num_users(self) -> int:
        """The number of users parsed in response."""
        return len(self._users)

    @error_check_output
    def _parse_users(self) -> List[User]:
        """
        Parse each individual user detail from response.

        Returns:
            List[User]: A list of User classes, one for each user detected.
        """
        return [
            User(**entry) for entry in self.extract_objects(
                schema=self._schema
            )
        ]


nodes_error_api = {
    'entry': {'errors': {'type': 'list'}},
    'objects': {
        'code': {'code': {'type': 'int'}},
        'message': {'message': {'type': 'str'}}
    }
}
nodes_list_add_member = {
    'entry': {'list': {'type': 'dict', 'children': {
        'user_results': {'type': 'dict', 'children': {
            'result': {'type': 'dict'}}}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}
nodes_list_remove_member = {
    'entry': {'list': {'type': 'dict', 'children': {
        'user_results': {'type': 'dict', 'children': {
            'result': {'type': 'dict'}}}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}
nodes_create_tweet = {
    'entry': {'create_tweet': {'type': 'dict', 'children': {
        'tweet_results': {'type': 'dict', 'children': {
            'result': {'type': 'dict'}}}}}
    },
    'objects': {
        'user': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict'}}}}}
        },
        'user_info': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict', 'children': {
                    'legacy': {'type': 'dict'}}}}}}}
        },
        'tweet': {'legacy': {'type': 'dict'}},
        'source': {'source': {'type': 'str'}}
    }
}
nodes_list_create = {
    'entry': {'list': {'type': 'dict'}},
    'objects': {
        'name': {'name': {'type': 'str'}},
        'description': {'description': {'type': 'str'}},
        'is_following': {'following': {'type': 'bool'}},
        'list_id': {'id_str': {'type': 'str'}},
        'member_count': {'member_count': {'type': 'int'}},
        'mode': {'mode': {'type': 'str'}},
    }
}
nodes_list_latest_tweets = {
    'entry': {'list': {'type': 'dict', 'children': {
        'tweets_timeline': {'type': 'dict', 'children': {
            'timeline': {'type': 'dict', 'children': {
                'instructions': {'type': 'list', 'children': {
                    'entries': {'type': 'list', 'children': {
                        'content': {'type': 'dict', 'children': {
                            'itemContent': {'type': 'dict', 'children': {
                                'tweet_results': {'type': 'dict', 'children': {
                                    'result': {'type': 'dict'}}}}}}}}}}}}}}}}}
    },
    'objects': {
        'user': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict'}}}}}
        },
        'user_info': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict', 'children': {
                    'legacy': {'type': 'dict'}}}}}}}
        },
        'tweet': {'legacy': {'type': 'dict'}},
        'source': {'source': {'type': 'str'}}
    }
}
nodes_user_tweets = {
    'entry': {'user': {'type': 'dict', 'children': {
        'result': {'type': 'dict', 'children': {
            'timeline_v2': {'type': 'dict', 'children': {
                'timeline': {'type': 'dict', 'children': {
                    'instructions': {'type': 'list', 'children': {
                        'entries': {'type': 'list', 'children': {
                            'content': {'type': 'dict', 'children': {
                                'itemContent': {'type': 'dict', 'children': {
                                    'tweet_results': {'type': 'dict', 'children': {
                                        'result': {'type': 'dict'}}}}}}}}}}}}}}}}}}}
    },
    'objects': {
        'user': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict'}}}}}
        },
        'user_info': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict', 'children': {
                    'legacy': {'type': 'dict'}}}}}}}
        },
        'tweet': {'legacy': {'type': 'dict'}},
        'source': {'source': {'type': 'str'}}
    }
}
nodes_tweet_result_by_id = {
    'entry': {'tweetResult': {'type': 'dict', 'children': {
        'result': {'type': 'dict'}}}
    },
    'objects': {
        'user': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict'}}}}}
        },
        'user_info': {'core': {'type': 'dict', 'children': {
            'user_results': {'type': 'dict', 'children': {
                'result': {'type': 'dict', 'children': {
                    'legacy': {'type': 'dict'}}}}}}}
        },
        'tweet': {'legacy': {'type': 'dict'}},
        'source': {'source': {'type': 'str'}}
    }
}
nodes_following = {
    'entry': {'user': {'type': 'dict', 'children': {
        'result': {'type': 'dict', 'children': {
            'timeline': {'type': 'dict', 'children': {
                'timeline': {'type': 'dict', 'children': {
                    'instructions': {'type': 'list', 'children': {
                        'entries': {'type': 'list', 'children': {
                            'content': {'type': 'dict', 'children': {
                                'itemContent': {'type': 'dict', 'children': {
                                    'user_results': {'type': 'dict', 'children': {
                                        'result': {'type': 'dict'}}}}}}}}}}}}}}}}}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}
nodes_user_by_screen_name = {
    'entry': {'user': {'type': 'dict', 'children': {
        'result': {'type': 'dict'}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}
nodes_list_members = {
    'entry': {'list': {'type': 'dict', 'children': {
        'members_timeline': {'type': 'dict', 'children': {
            'timeline': {'type': 'dict', 'children': {
                'instructions': {'type': 'list', 'children': {
                    'entries': {'type': 'list', 'children': {
                        'content': {'type': 'dict', 'children': {
                            'itemContent': {'type': 'dict', 'children': {
                                'user_results': {'type': 'dict', 'children': {
                                    'result': {'type': 'dict'}}}}}}}}}}}}}}}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}
nodes_user_by_rest_id = {
    'entry': {'user': {'type': 'dict', 'children': {
        'result': {'type': 'dict'}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}
nodes_users_by_rest_ids = {
    'entry': {'users': {'type': 'list', 'children': {
        'result': {'type': 'dict'}}}
    },
    'objects': {
        'user': {'rest_id': {'type': 'str'}},
        'user_info': {'legacy': {'type': 'dict'}}
    }
}


import logging
from logging import Logger

def logging_config(logger: Logger) -> None:
    """Create logging configuration."""
    stream_handler = logging.StreamHandler()
    stream_handler.setLevel(logging.DEBUG)

    formatter = logging.Formatter('%(levelname)s - %(message)s')
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)